/*
 * image_io.c --- This is the image implementation
 *  of the I/O manager.
 *
 * It operates on an EXT2 file system in an image file.
 *
 *
 */

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include <stdio.h>
#include <string.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#include <fcntl.h>
#include <time.h>

#include <ext2fs/ext2_fs.h>
#include <ext2fs/ext2fs.h>
#include "image_io.h"

#ifdef EXT2_MAC_OS_X
#define USE_MAC_FILE_MANAGER 1
#else
#define USE_MAC_FILE_MANAGER 0
#endif


#if USE_MAC_FILE_MANAGER
#include <CoreServices/CoreServices.h>
#endif

/*
 * For checking structure magic numbers...
 */

#define EXT2_CHECK_MAGIC(struct, code) \
    if ((struct)->magic != (code)) return (code)

struct image_private_data
{
    int magic;
    ext2_loff_t offset;

#if USE_MAC_FILE_MANAGER
    short ioRef;
#else
    FILE *image;
#endif
};


static errcode_t image_open(const char *name, int flags, io_channel *channel);
static errcode_t image_close(io_channel channel);
static errcode_t image_set_blksize(io_channel channel, int blksize);
static errcode_t image_read_blk(io_channel channel, unsigned long block,
                                int count, void *data);
static errcode_t image_write_blk(io_channel channel, unsigned long block,
                                 int count, const void *data);
static errcode_t image_flush(io_channel channel);
static errcode_t image_write_byte(io_channel channel, unsigned long offset,
                                  int size, const void *data);
static errcode_t image_set_option(io_channel channel, const char *option,
                                  const char *arg);



static struct struct_io_manager struct_image_manager =
    {
        EXT2_ET_MAGIC_IO_MANAGER,
        "Image I/O Manager",
        image_open,
        image_close,
        image_set_blksize,
        image_read_blk,
        image_write_blk,
        image_flush,
        image_write_byte,
        image_set_option
    };

EXT2FS io_manager image_io_manager = &struct_image_manager;


static errcode_t image_open(const char *name, int flags, io_channel *channel)
{
    io_channel io = NULL;
    struct image_private_data *data = NULL;
    errcode_t retval;
    int  open_flags;

    if (name == 0)
    {
        return EXT2_ET_BAD_DEVICE_NAME;
    }
    retval = ext2fs_get_mem(sizeof(struct struct_io_channel), &io);
    if (retval)
    {
        return retval;
    }
    memset(io, 0, sizeof(struct struct_io_channel));
    io->magic = EXT2_ET_MAGIC_IO_CHANNEL;
    retval = ext2fs_get_mem(sizeof(struct image_private_data), &data);
    if (retval)
    {
        goto cleanup;
    }

    io->manager = image_io_manager;
    retval = ext2fs_get_mem(strlen(name)+1, &io->name);
    if (retval)
    {
        goto cleanup;
    }

    strcpy(io->name, name);
    io->private_data = data;
    io->block_size = 1024;
    io->read_error = 0;
    io->write_error = 0;
    io->refcount = 1;

    memset(data, 0, sizeof(struct image_private_data));
    data->magic = EXT2_ET_MAGIC_UNIX_IO_CHANNEL;

    open_flags = (flags & IO_FLAG_RW) ? O_RDWR : O_RDONLY;
    if (flags & IO_FLAG_EXCLUSIVE)
    {
        open_flags |= O_EXCL;
    }

#if USE_MAC_FILE_MANAGER
    FSRef theRef;
    OSStatus status;

    FSPathMakeRef((const UInt8 *)name, &theRef, NULL);
    HFSUniStr255 dfName;
    FSGetDataForkName(&dfName);
    status = FSOpenFork(&theRef, dfName.length, dfName.unicode, fsRdWrPerm, &data->ioRef);
    if (status != noErr)
    {
        retval = EXT2_ET_BAD_DEVICE_NAME;
        goto cleanup;
    }
#else
    /* open the file */
    data->image = fopen(name, "r+b");

    if (data->image == 0)
    {
        retval = errno;
        goto cleanup;
    }
#endif

    *channel = io;
    return 0;

cleanup:
    if (data)
    {
        ext2fs_free_mem(&data);
    }
    if (io)
        ext2fs_free_mem(&io);
    return retval;
}

static errcode_t image_close(io_channel channel)
{
    struct image_private_data *data;
    errcode_t retval = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data *) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    if (--channel->refcount > 0)
    {
        return 0;
    }

#if USE_MAC_FILE_MANAGER
    if (noErr != FSCloseFork(data->ioRef))
#else
    if (fclose(data->image) < 0)
#endif
    {
        retval = errno;
    }

    ext2fs_free_mem(&channel->private_data);
    if (channel->name)
    {
        ext2fs_free_mem(&channel->name);
    }
    ext2fs_free_mem(&channel);
    return retval;

}

static errcode_t image_set_blksize(io_channel channel, int blksize)
{
    struct image_private_data *data;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data *) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    if (channel->block_size != blksize)
    {
        channel->block_size = blksize;
    }
    return 0;
}

static errcode_t image_read_blk(io_channel channel, unsigned long block,
                                int count, void *buf)
{
    struct image_private_data *data;
    errcode_t retval;
    size_t  size;
    ext2_loff_t location;
    int  actual = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data *) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    size = (count < 0) ? -count : count * channel->block_size;
    location = ((ext2_loff_t) block * channel->block_size) + data->offset;
#if USE_MAC_FILE_MANAGER
    if (noErr != FSSetForkPosition(data->ioRef, fsFromStart, location))
#else
    if (fseek(data->image, location, SEEK_SET) != 0)
#endif
    {
        retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
        goto error_out;
    }
#if USE_MAC_FILE_MANAGER
    ByteCount actualCount = 0;
    FSReadFork(data->ioRef, fsAtMark, 0, size, buf, &actualCount);
    actual = actualCount;
#else
    actual = fread(buf, 1, size, data->image);
#endif
    if (actual != size)
    {
        if (actual < 0)
        {
            actual = 0;
        }
        retval = EXT2_ET_SHORT_READ;
        goto error_out;
    }
    return 0;

error_out:
    memset((char *) buf+actual, 0, size-actual);
    if (channel->read_error)
    {
        retval = (channel->read_error)(channel, block, count, buf,
                                       size, actual, retval);
    }
    return retval;

}

static errcode_t image_write_blk(io_channel channel, unsigned long block,
                                 int count, const void *buf)
{
    size_t  size;
    ext2_loff_t location;
    int  actual = 0;
    errcode_t retval;
    struct image_private_data *data;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data *) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

    if (count == 1)
    {
        size = channel->block_size;
    }
    else
    {
        if (count < 0)
        {
            size = -count;
        }
        else
        {
            size = count * channel->block_size;
        }
    }

    location = ((ext2_loff_t) block * channel->block_size) + data->offset;
#if USE_MAC_FILE_MANAGER
    if (noErr != FSSetForkPosition(data->ioRef, fsFromStart, location))
#else
    if (fseek(data->image, location, SEEK_SET) != 0)
#endif
    {
        retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
        goto error_out;
    }

#if USE_MAC_FILE_MANAGER
    ByteCount actualCount = 0;
    FSWriteFork(data->ioRef, fsAtMark, 0, size, buf, &actualCount);
    actual = actualCount;
#else
    actual = fwrite(buf, 1, size, data->image);
#endif
    if (actual != size)
    {
        retval = EXT2_ET_SHORT_WRITE;
        goto error_out;
    }
    return 0;

error_out:
    if (channel->write_error)
    {
        retval = (channel->write_error)(channel, block, count, buf,
                                        size, actual, retval);
    }

    return retval;

}

static errcode_t image_flush(io_channel channel)
{
    struct image_private_data *data;
    errcode_t retval = 0;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data *) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

#if USE_MAC_FILE_MANAGER
    FSFlushFork(data->ioRef);
#else
    fflush(data->image);
#endif
    return retval;
}

static errcode_t image_write_byte(io_channel channel, unsigned long offset,
                                  int size, const void *buf)
{
    struct image_private_data *data;
    size_t  actual;

    EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
    data = (struct image_private_data *) channel->private_data;
    EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

#if USE_MAC_FILE_MANAGER
    if (noErr != FSSetForkPosition(data->ioRef, fsFromStart, offset + data->offset))
#else
    if (fseek(data->image, offset + data->offset, SEEK_SET) != 0)
#endif
    {
        return errno;
    }

#if USE_MAC_FILE_MANAGER
    ByteCount actualCount = 0;
    FSWriteFork(data->ioRef, fsAtMark, 0, size, buf, &actualCount);
    actual = actualCount;
#else
    actual = fwrite(buf, 1, size, data->image);
#endif
    if (actual != size)
    {
        return EXT2_ET_SHORT_WRITE;
    }
    return 0;

}

static errcode_t image_set_option(io_channel channel, const char *option,
                                  const char *arg)
{
    return EXT2_ET_INVALID_ARGUMENT;
}
